/*
   +----------------------------------------------------------------------+
   | HipHop for PHP                                                       |
   +----------------------------------------------------------------------+
   | Copyright (c) 2010-present Facebook, Inc. (http://www.facebook.com)  |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
*/

#include "hphp/runtime/vm/record.h"

#include "hphp/runtime/base/autoload-handler.h"
#include "hphp/runtime/base/runtime-error.h"
#include "hphp/runtime/vm/frame-restore.h"
#include "hphp/runtime/vm/named-entity-defs.h"
#include "hphp/runtime/vm/preclass.h"
#include "hphp/runtime/vm/treadmill.h"
#include "hphp/runtime/vm/unit.h"

namespace HPHP {

Mutex g_recordsMutex;

PreRecordDesc::PreRecordDesc(Unit* unit, int line1, int line2,
                             const StringData* name, Attr attrs,
                             const StringData* parentName,
                             const StringData* docComment, Id id)
  : m_unit(unit)
  , m_namedEntity(NamedEntity::get(name))
  , m_line1(line1)
  , m_line2(line2)
  , m_id(id)
  , m_attrs(attrs)
  , m_name(name)
  , m_parentName(parentName)
  , m_docComment(docComment) {}

void PreRecordDesc::atomicRelease() {
  delete this;
}

PreRecordDesc::Field::Field(PreRecordDesc* record,
                            const StringData* name,
                            Attr attrs,
                            const StringData* userType,
                            const TypeConstraint& typeConstraint,
                            const StringData* docComment,
                            const TypedValue& val,
                            RepoAuthType repoAuthType,
                            UserAttributeMap userAttributes)
  : m_name(name)
  , m_mangledName(mangleFieldName(record->name(), name, attrs))
  , m_attrs(attrs)
  , m_userType{userType}
  , m_docComment(docComment)
  , m_val(val)
  , m_repoAuthType{repoAuthType}
  , m_typeConstraint{typeConstraint}
  , m_userAttributes(userAttributes)
{}

const StringData* PreRecordDesc::mangleFieldName(
    const StringData* recordName,
    const StringData* fieldName,
    Attr attrs) {
  return PreClass::manglePropName(recordName, fieldName, attrs);
}

void PreRecordDesc::checkDefaultValueType(
    const PreRecordDesc::Field& field) const {
  auto const& val = field.val();
  if (val.m_type == KindOfUninit) return;
  auto const& tc = field.typeConstraint();
  if (tc.isCheckable()) tc.verifyRecField(&val, m_name, field.name());
}


size_t RecordDesc::stableHash() const {
  return folly::hash::hash_combine(
    name()->hashStatic(),
    unit()->sn()
  );
}

void RecordDesc::atomicRelease() {
  assertx(!m_cachedRecordDesc.bound());
  assertx(!getCount());
  this->~RecordDesc();
  lower_free(this);
}

void RecordDesc::setCached() {
  m_cachedRecordDesc.initWith(this);
}

void RecordDesc::setRecordDescHandle(
    rds::Link<LowPtr<RecordDesc>, rds::Mode::NonLocal> link) const {
  assertx(!m_cachedRecordDesc.bound());
  m_cachedRecordDesc = link;
}

bool RecordDesc::verifyPersistent() const {
  if (!isPersistent()) return false;
  assertx(!m_parent || recordHasPersistentRDS(m_parent.get()));
  return true;
}

void RecordDesc::destroy() {
  if (!m_cachedRecordDesc.bound()) return;

  Lock l(g_recordsMutex);
  if (!m_cachedRecordDesc.bound()) return;
  m_cachedRecordDesc = rds::Link<LowPtr<RecordDesc>, rds::Mode::NonLocal>{};
  m_preRec->namedEntity()->removeRecordDesc(this);
  m_parent.reset();

  Treadmill::enqueue(
    [this] {
      if (!this->decAtomicCount()) this->atomicRelease();
    }
  );
}

Slot RecordDesc::lookupField(const StringData* fieldName) const {
  return m_fields.findIndex(fieldName);
}

const RecordDesc::Field& RecordDesc::field(Slot idx) const {
  assertx(idx != kInvalidSlot && idx < numFields());
  return m_fields[idx];
}

RecordDesc::Avail RecordDesc::availWithParent(
    RecordDesc*& parent, bool tryAutoload /* = false */) const {
  if (RecordDesc* ourParent = m_parent.get()) {
    if (!parent) {
      auto pprec = ourParent->m_preRec.get();
      parent = get(pprec->namedEntity(),
                                   parentName(), tryAutoload);
      if (!parent) {
        parent = ourParent;
        return Avail::Fail;
      }
    }
    if (parent != ourParent) {
      if (UNLIKELY(ourParent->isZombie())) {
        const_cast<RecordDesc*>(this)->destroy();
      }
      return Avail::False;
    }
  }
  return Avail::True;
}

RecordDesc* RecordDesc::newRecordDesc(PreRecordDesc* preRec,
                                      RecordDesc* parent) {
  auto const mem = lower_malloc(sizeof(RecordDesc));
  try {
    return new (mem) RecordDesc(preRec, parent);
  } catch (...) {
    lower_free(mem);
    throw;
  }
}

void RecordDesc::setParent() {
  // validate the parent
  if (m_parent.get() != nullptr) {
    auto parentAttrs = m_parent->attrs();
    if (UNLIKELY(parentAttrs & AttrFinal)) {
      raise_error("Record %s may not inherit from non-abstract record %s",
                  name()->data(), m_parent->name()->data());
    }
  }
}

void RecordDesc::setFields() {
  FieldMap::Builder curFieldMap;
  if (m_parent.get() != nullptr) {
    for (auto const& parentField : m_parent->allFields()) {
      curFieldMap.add(parentField.name(), parentField);
    }
  }
  for (auto const& preField : m_preRec->allFields()) {
    auto const it = curFieldMap.find(preField.name());
    if (it != curFieldMap.end()) {
      raise_error("Cannot redeclare record field %s::%s",
                  name()->data(),
                  preField.name()->data());
    }
    m_preRec->checkDefaultValueType(preField);
    curFieldMap.add(preField.name(), preField);
  }
  m_fields.create(curFieldMap);
}

RecordDesc::RecordDesc(PreRecordDesc* preRec, RecordDesc* parent)
  : m_parent(parent)
  , m_preRec(PreRecordDescPtr(preRec)) {
  setParent();
  setFields();
}

bool RecordDesc::recordDescOf(const RecordDesc* rec) const {
  // TODO: Optimize this. See T45403957.
  assertx(rec);
  auto curr = this;
  while(curr) {
    if (curr == rec) return true;
    curr = curr->m_parent.get();
  }
  return false;
}

const RecordDesc* RecordDesc::commonAncestor(const RecordDesc* rec) const {
  // TODO: Optimize this. See T45403957.
  assertx(rec);
  std::vector<const RecordDesc*> thisVec;
  std::vector<const RecordDesc*> recVec;
  for (auto r = this; r != nullptr; r = r->m_parent.get()) thisVec.push_back(r);
  for (auto r = rec; r != nullptr; r = r->m_parent.get()) recVec.push_back(r);
  auto idx = std::min(thisVec.size(), recVec.size()) - 1;
  do {
    if (thisVec[idx] == recVec[idx]) return thisVec[idx];
  } while(idx--);
  return nullptr;
}

namespace {

void setupRecord(RecordDesc* newRecord, NamedEntity* nameList) {
  bool const isPersistent = newRecord->isPersistent();
  assertx(!isPersistent || newRecord->verifyPersistent());
  nameList->m_cachedRecordDesc.bind(
    isPersistent? rds::Mode::Persistent : rds::Mode::Normal,
    rds::LinkName{"NERecord", newRecord->name()}
  );
  newRecord->setRecordDescHandle(nameList->m_cachedRecordDesc);
  newRecord->incAtomicCount();
  nameList->pushRecordDesc(newRecord);
}

}

RecordDesc* RecordDesc::def(PreRecordDesc* preRecord,
                            bool failIsFatal /* = true */) {
  auto const nameList = preRecord->namedEntity();

  // Error out if there is already a different type
  // with the same name in the request
  auto existingKind = nameList->checkSameName<RecordDesc>();
  if (existingKind) {
    FrameRestore fr(preRecord);
    raise_error("Cannot declare record with the same (%s) as an "
                "existing %s", preRecord->name()->data(), existingKind);
    return nullptr;
  }

  // If there was already a record declared with DefRecordDesc, check if it's
  // compatible.
  if (auto cachedRec = nameList->getCachedRecordDesc()) {
    if (cachedRec->preRecordDesc() != preRecord) {
      if (failIsFatal) {
        FrameRestore fr(preRecord);
        raise_error("Record already declared: %s", preRecord->name()->data());
      }
      return nullptr;
    }
    return cachedRec;
  }

  // Get a compatible predefined record, if one exists. Otherwise, add the
  // current one to the list of defined records.
  // Set the cached record in either case.
  RecordDesc* parent = nullptr;
  auto top = nameList->recordList();
  for (;;) {
    for (auto rec = top; rec != nullptr; rec = rec->m_next) {
      if (rec->preRecordDesc() != preRecord) continue;
      auto avail = rec->availWithParent(parent, failIsFatal /*tryAutoload*/);
      if (LIKELY(avail == RecordDesc::Avail::True)) {
        rec->setCached();
        return rec;
      }
      if (avail == RecordDesc::Avail::Fail) {
        // parent is not available and cannot be autoloaded
        if (failIsFatal) {
          FrameRestore fr(preRecord);
          raise_error("unknown record %s", parent->name()->data());
        }
        return nullptr;
      }
      assertx(avail == RecordDesc::Avail::False);
    }

    // Create a new record.
    if (!parent && preRecord->parentName()->size() != 0) {
      // Load the parent
      parent = get(preRecord->parentName(), failIsFatal);
      if (parent == nullptr) {
        if (failIsFatal) {
          FrameRestore fr(preRecord);
          raise_error("unknown record %s", preRecord->parentName()->data());
        }
        return nullptr;
      }
    }

    RecordDescPtr newRecord;
    {
      FrameRestore fr(preRecord);
      newRecord = RecordDesc::newRecordDesc(
          const_cast<PreRecordDesc*>(preRecord), parent);
    }

    Lock l(g_recordsMutex);

    if (UNLIKELY(top != nameList->recordList())) {
      top = nameList->recordList();
      continue;
    }

    setupRecord(newRecord.get(), nameList);

    newRecord->setCached();
    return newRecord.get();
  }
}

RecordDesc* RecordDesc::loadMissing(const NamedEntity* ne,
                                    const StringData* name) {
  VMRegAnchor _;
  AutoloadHandler::s_instance->autoloadRecordDesc(
    StrNR(const_cast<StringData*>(name)));
  return lookup(ne);
}

RecordDesc* RecordDesc::get(const StringData* name, bool tryAutoload) {
  String normStr;
  auto ne = NamedEntity::get(name, true, &normStr);
  if (normStr) {
    name = normStr.get();
  }
  return get(ne, name, tryAutoload);
}

RecordDesc* RecordDesc::get(const NamedEntity* ne,
                            const StringData *name, bool tryAutoload) {
  RecordDesc *rec = lookup(ne);
  if (UNLIKELY(!rec)) {
    if (tryAutoload) {
      return loadMissing(ne, name);
    }
  }
  return rec;
}

}
